import { Group } from 'spritejs'
import { isArray, clone } from '../util'
import { BaseNode } from './BaseNode'
import InteractionHooks from './InteractionHooks'

const _colors = Symbol('colors')
const _shouldUseBuiltInColors = Symbol('shouldUseBuiltInColors')
const filterEmpty = data => (isArray(data) ? data.filter(Boolean) : data)
const observe = (obj, callback) => {
  return new Proxy(obj, {
    get(target, key, receiver) {
      return Reflect.get(target, key, receiver)
    },

    set(target, key, value, receiver) {
      if (key === 'length') {
        if (typeof callback === 'function') {
          callback(obj)
        }
      }

      return Reflect.set(target, key, value, receiver)
    }
  })
}

export class BaseVisual extends BaseNode {
  constructor(attrs = {}) {
    super(attrs)
    this.container = null
    this.data = null
    this.dataOrigin = null
    this.dataFields = {}
    this.schema = {}
    this[_colors] = []
    this[_shouldUseBuiltInColors] = true
    this.chart = null
  }

  connectToChart(chart) {
    this.chart = chart
    this.$group = new Group({ zIndex: 2 }) //  插件grid的zIndex为1
    this.$group.attr({ clipOverflow: false, size: [0.01, 0.01] }) // 设置group不裁切内部元素，并且设置大小强制渲染
    this.chart.$group.appendChild(this.$group)

    if (this[_shouldUseBuiltInColors]) {
      this[_colors] = this.chart.getStyle('colors')()
    }

    this.afterConnectedToChart()
  }

  afterConnectedToChart() {
    this.listenTo(this.chart, InteractionHooks.changeTheme, this.onThemeChange)
    this.listenTo(
      this.chart,
      InteractionHooks.toggleLegend,
      this.onLegendToggle
    )
  }

  onThemeChange() {
    if (this[_shouldUseBuiltInColors]) {
      this[_colors] = this.chart.getStyle('colors')()
    }

    this.update(this.data, this.data) // 直接更新所有元素（DataDiffer 对相同元素是做 update 操作）
  }

  // 监听 Legend 变化
  onLegendToggle(evt) {
    const { index, disabled } = evt
    this.data[index].disabled = disabled
    this.update(this.dataOrigin, this.data)
  }

  source(data = []) {
    if (!isArray(data)) {
      data = [data]
    }

    data = filterEmpty(data)

    this.data = observe(data, data => {
      const oldData = this.dataOrigin
      this.update(oldData, filterEmpty(data))
      this.dataOrigin = filterEmpty(clone(data))
    })

    if (this.dataOrigin) {
      // 图表已经渲染，这时候重新设置数据，图表应当更新即可
      this.update(this.dataOrigin, this.data)
    }

    this.dataOrigin = filterEmpty(clone(data)) // 保存一份数据快照

    return this
  }

  setDataFields(dataFields) {
    this.dataFields = dataFields

    return this
  }

  color(i) {
    return this[_colors][i] || this[_colors][0] || '#f00'
  }

  useColor(colors) {
    if (!isArray(colors)) {
      colors = [colors]
    }

    this[_colors] = colors
    this[_shouldUseBuiltInColors] = false // 此时就不应该再使用内置主题配色
    return this
  }

  render() {
    throw new Error(`The render method must be implemented!`)
  }

  update() {
    throw new Error(`The update method must be implemented!`)
  }

  forceUpdate() {
    if (!this.chart) {
      return
    }

    this.chart.$group.removeChild(this.$group)
    this.$group = new Group()
    this.chart.$group.appendChild(this.$group)
    this.render()
  }

  toJSON() {
    return {
      attrs: this.getAttr(),
      style: this.getStyle()(),
      dataFields: this.dataFields,
      source: this.dataOrigin
    }
  }
}
