import { Group, Scene } from 'spritejs'
import { isObject, debounce } from '../util'
import ResizeObserver from './ResizeObserver'
import { Attr } from './Attr'

const $ = node =>
  typeof node === 'string' ? document.querySelector(node) : node

export class Plot {
  constructor(container, opts) {
    if (isObject(container)) {
      opts = container
      container = opts.container
    }

    if (!container) {
      throw new Error(
        `Need a dom node to render chart! Check if you set the container`
      )
    }

    this.attr = new Attr()
    this.attr.set(opts)
    this.parentNode = container // 挂载节点（有可能是 string，如『#app』）
    this.dom = $(container) // 挂载节点 DOMNode
    this.initScene(this.dom, opts)
    this.plots = []
    // this.initPlots()
    this.charts = []
  }

  /**
   * 初始化舞台
   * @param {*} container
   * @param {*} opts
   */
  initScene(container, opts) {
    // 接口统一改成
    // { forceFit: true } || { width: Number, height: Number } || { size: [Number, Number] }
    if (opts.forceFit) {
      opts.viewport = 'auto'
    } else {
      opts.viewport = opts.size ? opts.size : [opts.width, opts.height]
    }

    this.scene = new Scene(container, { displayRatio: 'auto', ...opts })
    this.layer = this.scene.layer()
    this.canvas = this.layer.canvas

    if (opts.forceFit) {
      this.forceFit()
    }
  }

  initPlots() {
    this.root = new Group({
      size: this.scene.viewport,
      boxSizing: 'border-box'
    }) // 根容器
    this.layer.appendChild(this.root)
    this.plots = [
      { container: this.root, x: 0, y: 0, width: '100%', height: '100%' }
    ] // 绘制区域
  }

  forceFit() {
    const dom = this.dom
    let isFirstRender = true // 首次渲染时也会触发 resize

    const observer = new ResizeObserver(
      debounce(entries => {
        entries.forEach(entry => {
          if (!isFirstRender) {
            const { width, height } = entry.contentRect
            this.onResize(width, height)
          } else {
            isFirstRender = false
          }
        })
      }, 300)
    )

    observer.observe(dom)
  }

  onResize(w, h) {
    // FIXME: 必须同时修改才生效，疑惑点，但是这样做的结果看起来是对的。
    this.scene.setViewport(w, h)
    this.scene.setResolution(w, h)

    this.plots.forEach(({ container, x, y, width, height }) => {
      container.attr(this.calcPlotPosAndSize(x, y, width, height))
    })

    this.charts.forEach(chart => chart.forceUpdate())
  }

  calcPlotPosAndSize(x, y, width, height) {
    const viewport = this.scene.viewport
    const getPercentFromString = str => parseFloat(str) / 100
    const [posX, w] = [x, width].map(n => {
      return typeof n === 'string' ? getPercentFromString(n) * viewport[0] : n
    })
    const [posY, h] = [y, height].map(n => {
      return typeof n === 'string' ? getPercentFromString(n) * viewport[1] : n
    })
    return { pos: [posX, posY], size: [w, h] }
  }

  subPlot(x, y, width, height) {
    const plot = new Group({
      ...this.calcPlotPosAndSize(x, y, width, height),
      boxSizing: 'border-box'
    }) // 更新到指定的绘制区域
    this.layer.appendChild(plot)
    this.plots.push({ container: plot, x, y, width, height })
    return plot
  }

  addChart(chart) {
    chart.id = this.plots.length + 1 // 设置 chart 的 ID
    this.charts.push(chart)
  }

  setCanvasCursor(type = 'default') {
    this.canvas.style.cursor = type
  }
}
