import { isObject } from '../../util'
import { BaseNode } from '../BaseNode'
import { Plot } from '../Plot'
import InteractionHooks from '../InteractionHooks'
import lifecycle from './lifecycle'
import { Title } from './Title'
import { PlotInstancesMap, VisualsMap, ThemeMap } from './registry'

const $ = node =>
  typeof node === 'string' ? document.querySelector(node) : node

const defaultTheme = 'light'
const _title = Symbol('title')
const _visuals = Symbol('visuals')
const _plugins = Symbol('plugins')
const _isRendered = Symbol('isRendered')
const _initScene = Symbol('initScene')
const _mountToPlot = Symbol('mountToPlot')
const _delegateHooksEvent = Symbol('delegateHooksEvent')
const _renderTitle = Symbol('renderTitle')
export const _addVisual = Symbol('addVisual')

export class Chart extends BaseNode {
  static supportedVisuals() {
    const types = []
    for (let type of VisualsMap.keys()) {
      types.push(type)
    }
    return types
  }

  constructor(opts = {}) {
    if (!isObject(opts)) {
      throw new Error(`Need a dom node to render chart!`)
    }

    const defaultAttrs = {
      pos: [0, 0],
      size: ['100%', '100%'],
      forceFit: false
    }

    const attrs = Object.assign({}, defaultAttrs, opts)

    super(attrs)
    this.id = null
    this.canvas = null
    this.$group = null
    this[_isRendered] = false

    this[_initScene](attrs)
    this.lifecycle = lifecycle
    this.emit(this.lifecycle.beforeCreate, this)

    this.setTheme(defaultTheme) // defaultTheme
    this.title = null // 实例化后的 Title
    this[_title] = attrs.title || null // 图表标题原始数据
    this[_visuals] = []
    this[_plugins] = []

    this.emit(this.lifecycle.created)

    this[_mountToPlot]()
    this[_renderTitle]()
  }

  supportedVisuals() {
    return Chart.supportedVisuals()
  }

  [_initScene](attrs) {
    this.dom = $(attrs.container)
    let plot = PlotInstancesMap.get(this.dom)

    if (!plot) {
      plot = new Plot(attrs)
      PlotInstancesMap.set(this.dom, plot)
    }

    this.plot = plot
    this.canvas = plot.canvas
    this.plot.addChart(this)
  }

  [_mountToPlot]() {
    const [x, y] = this.attr.get('pos')
    const [width, height] = this.attr.get('size')
    const group = this.plot.subPlot(x, y, width, height)
    this.$group = group
    this.attr.set('size', this.$group.attr('size'))
    this.attr.set('pos', this.$group.attr('pos'))
  }

  [_renderTitle]() {
    const title = new Title(this.getTitle())
    title.connect(this)
    title.render()
    this.title = title
  }

  setTheme(name) {
    const target = ThemeMap.get(name)
    if (!target) return
    Object.keys(target).forEach(key => this.useStyle(key, target[key]))
    this[_isRendered] && this.emit(InteractionHooks.changeTheme, target)
    this.onThemeChange()
  }

  onThemeChange() {
    if (!this.$group || !this.title) {
      return
    }
    this.$group.attr(this.getStyle('Chart')() || {})
    this.title.update(this.getTitle())
  }

  getTitle() {
    return { text: this[_title], ...(this.getStyle('Title')() || {}) }
  }

  setTitle(title, style = {}) {
    if (!title) return
    this[_title] = title
    typeof style === 'object' && this.useStyle('Title', style)
    this.title && this.title.update(this.getTitle())
  }

  addPlugin(plugin) {
    plugin.connectToChart && plugin.connectToChart(this)
    this[_plugins].push(plugin)
    return this
  }

  [_addVisual](type, attrs) {
    if (!VisualsMap.get(type)) {
      throw new Error(
        `Unknown visual type.
        you can use Chart.supportedVisuals() to know which visual is supported!`
      )
    }

    const Visual = VisualsMap.get(type)
    const visual = new Visual(attrs)
    visual.connectToChart && visual.connectToChart(this)
    this[_visuals].push(visual)
    return visual
  }

  [_delegateHooksEvent]() {
    const children = [...this[_visuals], ...this[_plugins]]

    Object.keys(InteractionHooks).forEach(key => {
      children.forEach(child => {
        this.listenTo(child, InteractionHooks[key], (...args) => {
          this.emit(InteractionHooks[key], ...args)
        })
      })
    })

    // void [InteractionHooks.showTooltip].forEach(type => {
    //   this.on(type, () => this.setCanvasCursor('pointer'))
    // })
    // void [InteractionHooks.hideTooltip].forEach(type => {
    //   this.on(type, () => this.setCanvasCursor('default'))
    // })
  }

  setCanvasCursor(type = 'default') {
    this.canvas.style.cursor = type
  }

  render() {
    this[_delegateHooksEvent]()
    this.emit(this.lifecycle.beforeRender, this)
    this[_visuals].forEach(visual => visual.render && visual.render())
    this[_plugins].forEach(plugin => plugin.render && plugin.render())
    this.emit(this.lifecycle.rendered, this)
    this[_isRendered] = true
  }

  forceUpdate() {
    this.attr.set('size', this.$group.attr('size'))
    this.attr.set('pos', this.$group.attr('pos'))
    this.emit(this.lifecycle.beforeUpdate, this)
    this[_visuals].forEach(visual => visual.forceUpdate && visual.forceUpdate())
    this[_plugins].forEach(plugin => plugin.forceUpdate && plugin.forceUpdate())
    this.emit(this.lifecycle.updated, this)
  }

  destroy() {
    this.emit(this.lifecycle.beforeDestroy, this)
    this.$group.parentNode.removeChild(this.$group)
    this.emit(this.lifecycle.destroyed, this)
  }
}
